from functools import partial

from kazoo.exceptions import NoNodeError
from kazoo.protocol.states import EventType

from distconfig.backends.base import BaseBackend


class ZooKeeperBackend(BaseBackend):
    """Zooekeeper backend implentation.

    If you are using gevent, make sure that the kazoo client is setup to
    use gevent event handler e.g. ``KazooClient(..., handler=SequentialGeventHandler())``

    User must call ``KazooClient.start()`` before using the backend.

    :param client: Instance of :class:`kazoo.client.KazooClient`.

    """

    def __init__(self, client, **kwargs):
        super(ZooKeeperBackend, self).__init__(**kwargs)
        self._client = client

    def get_raw(self, path):
        try:
            return self._get_and_watch_path(path)
        except NoNodeError:
            return self._get_and_watch_unexistant_path(path)

    def _get_and_watch_unexistant_path(self, path):
        self._client.retry(self._client.exists, path, watch=partial(self._on_path_change, path))

    def _get_and_watch_path(self, path):
        data, _ = self._client.retry(self._client.get, path, watch=partial(self._on_path_change, path))
        return data

    def _on_path_change(self, path, event):
        # XXX: Before we could set the new watches the data may have changed already
        # in ZooKeeper backend, in that case we may have missed the notification.
        # AFAIK there is no easy way to work around this.
        # https://zookeeper.apache.org/doc/r3.1.2/zookeeperProgrammers.html#sc_WatchRememberThese
        if event.type == EventType.DELETED:
            data = self._get_and_watch_unexistant_path(path)
        else:
            data = self._get_and_watch_path(path)
        return super(ZooKeeperBackend, self)._notify_listeners(data)



print("import: 'distconfig'")
import distconfig

print("import: 'distconfig.backends'")
import distconfig.backends

print("import: 'distconfig.tests'")
import distconfig.tests

print("import: 'distconfig.tests.integration'")
import distconfig.tests.integration

print("import: 'distconfig.tests.unit'")
import distconfig.tests.unit




import abc
import json
import logging
import sys


LOGGER = logging.getLogger(__name__)


class BaseBackend(metaclass=abc.ABCMeta):
    """Base abstract backend class.

    Backend implementation should inherit and implement ``get_raw`` method.

    :param parser: Callable that accept a string and parse it, default: ``json.loads``.
    :param logger: :class:`logging.Logger`` instance.
    """

    def __init__(self, parser=json.loads, logger=LOGGER):
        self.__callbacks = []
        self.__parser = parser

        self._logger = logger

    @abc.abstractmethod
    def get_raw(self, path):
        """Get path value from backend as it is.

        :path: key in the backend.
        :return: path value as saved in backend or None if not found in backend.
        """

    def get(self, path):
        """Get parsed path value in backend.

        :path: key in the backend.
        :return: path value parsed.
        """
        data = self.get_raw(path)
        return self._parse_raw_data(data)

    def _parse_raw_data(self, data):
        if data is None:
            return {}
        return self.__parser(data)

    def add_listener(self, callback):
        """Add callback to be called when data change in the backend.

        If the same callback is added more than once, then it will be notified more than once.
        That is, no check is made to ensure uniqueness.

        :param callback: Callable that accept one argument the new data.
        """
        self.__callbacks.append(callback)

    def remove_listener(self, callback):
        """Remove previously added callback.

        If callback had been added more than once, then only the first occurrence will be removed.

        :param callback: Callable as with ``:meth: add_listener``.
        :raise ValueError: In case callback was not previously registered.
        """
        self.__callbacks.remove(callback)

    def _notify_listeners(self, value):
        self._logger.debug('Notify listeners of new value: %r', value)
        value = self._parse_raw_data(value)
        last_exc = None
        for callback in self.__callbacks:
            try:
                callback(value)
            except Exception:
                last_exc = sys.exc_info()
        if last_exc:
            self._logger.exception('Notify listeners raised an exception', exc_info=last_exc)
            try:
                raise last_exc
            finally:
                del last_exc
